// Copyright (c) 2016-2017 <daggerbot@gmail.com>
// This software is available under the terms of the zlib license.
// See COPYING.md for more information.

use std;
use std::fmt::{self, Display, Formatter};
use std::ptr;

use kernel32;
use winapi;

use mem::LocalSlice;

/// Windows API error.
#[derive(Clone, Copy, Debug, Eq, Hash, PartialEq)]
pub struct Error {
    code: u32,
}

impl Error {
    pub fn code (self) -> u32 { self.code }

    pub fn get () -> Option<Error> {
        unsafe {
            Error::from_code(kernel32::GetLastError())
        }
    }

    pub fn from_code (code: u32) -> Option<Error> {
        match code {
            0 => None,
            _ => Some(Error { code: code }),
        }
    }

    pub fn format_msg_wide (self) -> Result<LocalSlice<u16>> {
        const FLAGS: u32 = winapi::FORMAT_MESSAGE_FROM_SYSTEM
                         | winapi::FORMAT_MESSAGE_ALLOCATE_BUFFER;

        let mut buf: *mut u16 = ptr::null_mut();

        unsafe {
            match kernel32::FormatMessageW(FLAGS, ptr::null(), self.code, 0,
                                           &mut buf as *mut *mut u16 as *mut u16, 0,
                                           ptr::null_mut())
            {
                0 => Err(Error::get().unwrap()),
                len => Ok(LocalSlice::from_raw_parts(buf, len as usize)),
            }
        }
    }

    pub fn msg_lossy (self) -> String {
        match self.format_msg_wide() {
            Ok(msg) => String::from_utf16_lossy(&msg[..]),
            Err(_) => format!("error code {}", self.code),
        }
    }

    pub fn try () -> Result<()> {
        match Error::get() {
            None => Ok(()),
            Some(err) => Err(err),
        }
    }
}

impl Display for Error {
    fn fmt (&self, f: &mut Formatter) -> fmt::Result {
        f.write_str(self.msg_lossy().as_str())
    }
}

impl std::error::Error for Error {
    fn description (&self) -> &str { "win32 error" }
}

/// Windows API result type.
pub type Result<T> = std::result::Result<T, Error>;

/// Sets the thread local error code to zero.
pub fn clear_error () {
    unsafe {
        kernel32::SetLastError(0);
    }
}

/// Sets the thread local error code.
pub fn set_error (error: Option<Error>) {
    let code = match error {
        None => 0,
        Some(error) => error.code,
    };

    unsafe {
        kernel32::SetLastError(code);
    }
}
