// Copyright (c) 2016-2017 <daggerbot@gmail.com>
// This software is available under the terms of the zlib license.
// See COPYING.md for more information.

use std::mem;
use std::ops::{Deref, DerefMut};
use std::slice;

use kernel32;

/// Boxed type managed with `LocalAlloc` and `LocalFree`.
pub struct Local<T: Sized> {
    ptr: *mut T,
}

impl<T> Local<T> {
    pub unsafe fn from_raw (ptr: *mut T) -> Local<T> { Local { ptr: ptr } }
}

impl<T> Deref for Local<T> {
    type Target = T;

    fn deref (&self) -> &T {
        unsafe {
            mem::transmute(self.ptr)
        }
    }
}

impl<T> DerefMut for Local<T> {
    fn deref_mut (&mut self) -> &mut T {
        unsafe {
            mem::transmute(self.ptr)
        }
    }
}

impl<T> Drop for Local<T> {
    fn drop (&mut self) {
        unsafe {
            if !kernel32::LocalFree(self.ptr as *mut _).is_null() {
                error!("LocalFree failed with error code {}", kernel32::GetLastError());
            }
        }
    }
}

/// Boxed type for slices managed with `LocalAlloc` and `LocalFree`.
pub struct LocalSlice<T: Sized> {
    ptr: *mut T,
    len: usize,
}

impl<T> LocalSlice<T> {
    pub unsafe fn from_raw_parts (ptr: *mut T, len: usize) -> LocalSlice<T> {
        LocalSlice {
            ptr: ptr,
            len: len,
        }
    }
}

impl<T> Deref for LocalSlice<T> {
    type Target = [T];

    fn deref (&self) -> &[T] {
        unsafe {
            slice::from_raw_parts(self.ptr as *const T, self.len)
        }
    }
}

impl<T> DerefMut for LocalSlice<T> {
    fn deref_mut (&mut self) -> &mut [T] {
        unsafe {
            slice::from_raw_parts_mut(self.ptr, self.len)
        }
    }
}

impl<T> Drop for LocalSlice<T> {
    fn drop (&mut self) {
        unsafe {
            if !kernel32::LocalFree(self.ptr as *mut _).is_null() {
                error!("LocalFree failed with error code {}", kernel32::GetLastError());
            }
        }
    }
}
